@@ -6,7 +6,7 @@
 # License           : BSD License
 # -----------------------------------------------------------------------------
 # Creation date     : 14-Jul-2013
-# Last modification : 30-Dec-2014
+# Last modification : 19-Jan-2015
 # -----------------------------------------------------------------------------
 
 import re, os, sys, argparse, json
@@ -17,7 +17,7 @@ try:
 except ImportError:
 	reporter = None
 
-VERSION = "0.0.8"
+VERSION = "0.1.0"
 LICENSE = "http://ffctn.com/doc/licenses/bsd"
 
 __doc__ = """
@@ -85,6 +85,7 @@ def grammar(g=None):
 	g.word    ("RP",               ")")
 	g.word    ("SELF",             "&")
 	g.word    ("COMMA",            ",")
+	g.word    ("MACRO",            "@macro")
 	g.word    ("SEMICOLON",        ";")
 	g.word    ("EQUAL",            "=")
 	g.word    ("LSBRACKET",        "[")
@@ -94,7 +95,7 @@ def grammar(g=None):
 	g.token   ("PERCENTAGE",       "\d+(\.\d+)?%")
 	g.token   ("STRING_SQ",        "'((\\\\'|[^'\\n])*)'")
 	g.token   ("STRING_DQ",        "\"((\\\\\"|[^\"\n])*)\"")
-	g.token   ("STRING_UQ",        "[^\s\n\*\+;\(\)\[\]]+")
+	g.token   ("STRING_UQ",        "[^\s\n\*\+,:;\(\)\[\]]+")
 	g.token   ("INFIX_OPERATOR",   "[\-\+\*\/]")
 
 	g.token   ("NODE",             "\*|([a-zA-Z][\-_a-zA-Z0-9\-]*)")
@@ -102,10 +103,11 @@ def grammar(g=None):
 	g.token   ("NODE_ID",          "#[_a-zA-Z][_a-zA-Z0-9\-]*")
 
 	# SEE: http://www.w3schools.com/cssref/css_units.asp
-	g.token   ("UNIT",             "em|ex|px|cm|mm|in|pt|pc|ch|rem|vh|vmin|vmax|s|deg|rad|grad|ms|Hz|kHz|\%")
-	g.token   ("VARIABLE_NAME",    "[\w_Processor()][\w\d_]*")
+	#g.token   ("UNIT",             "em|ex|px|pem|cm|mm|in|pt|pc|ch|rem|vh|vmin|vmax|s|deg|rad|grad|ms|Hz|kHz|\%")
+	g.token   ("UNIT",             "[\w]+|\%")
+	g.token   ("VARIABLE_NAME",    "[\w_][\w\d_]*")
 	g.token   ("METHOD_NAME",      "[\w_][\w\d_]*")
-	g.token   ("MACRO_NAME",       "[\w_][\w\d_]*")
+	g.token   ("NAME",             "[\w_][\w\d_]*")
 	g.token   ("REFERENCE",        "\$([\w_][\w\d_]*)")
 	g.token   ("COLOR_NAME",       "[a-z][a-z0-9\-]*")
 	g.token   ("COLOR_HEX",        "\#([A-Fa-f0-9][A-Fa-f0-9]?[A-Fa-f0-9]?[A-Fa-f0-9]?[A-Fa-f0-9]?[A-Fa-f0-9]?([A-Fa-f0-9][A-Fa-f0-9])?)")
@@ -141,8 +143,9 @@ def grammar(g=None):
 	g.group     ("Value",            s.Number, s.COLOR_HEX, s.COLOR_RGB, s.REFERENCE, s.String)
 	g.rule      ("Parameters",       s.VARIABLE_NAME, g.arule(s.COMMA, s.VARIABLE_NAME).zeroOrMore())
 	g.rule      ("Arguments",        s.Value, g.arule(s.COMMA, s.Value).zeroOrMore())
-
 	g.rule      ("Expression")
+	g.rule      ("ExpressionList", s.Expression._as("head"), g.arule(s.COMMA, s.Expression).zeroOrMore()._as("tail"))
+
 	# NOTE: We use Prefix and Suffix to avoid recursion, which creates a lot
 	# of problems with parsing expression grammars
 	g.group     ("Prefix", s.Value, g.arule(s.LP, s.Expression, s.RP))
@@ -150,6 +153,7 @@ def grammar(g=None):
 
 	g.rule      ("Invocation",   g.arule(s.DOT,     s.METHOD_NAME).optional()._as("method"), s.LP, s.Arguments.optional()._as("arguments"), s.RP)
 	g.rule      ("InfixOperation", s.INFIX_OPERATOR, s.Expression)
+	# TODO: Might be better to use COMMA as a suffix to chain expressions
 	s.Suffixes.set(s.InfixOperation, s.Invocation)
 
 	# =========================================================================
@@ -158,14 +162,15 @@ def grammar(g=None):
 
 	g.rule      ("Comment",          s.COMMENT.oneOrMore(), s.EOL)
 	g.rule      ("Include",          s.INCLUDE, s.PATH,     s.EOL)
-	g.rule      ("Declaration",      s.VARIABLE_NAME._as("name"), s.EQUAL, s.Expression._as("value"), s.EOL)
+	g.rule      ("Declaration",      s.SPECIAL_NAME.optional()._as("decorator"), s.VARIABLE_NAME._as("name"), s.EQUAL, s.ExpressionList._as("value"), s.EOL)
 
 	# =========================================================================
 	# OPERATIONS
 	# =========================================================================
 
-	g.rule      ("Assignment",       s.CSS_PROPERTY._as("name"), s.COLON, s.Expression.oneOrMore()._as("values"), s.IMPORTANT.optional()._as("important"), s.SEMICOLON.optional())
-	g.rule      ("MacroInvocation",  s.MACRO_NAME,   s.LP, s.Arguments.optional(), s.RP)
+	g.rule      ("Assignment",       s.CSS_PROPERTY._as("name"), s.COLON, s.ExpressionList.oneOrMore()._as("values"), s.IMPORTANT.optional()._as("important"), s.SEMICOLON.optional())
+	#g.rule      ("Assignment",       s.CSS_PROPERTY._as("name"), s.COLON, s.Expression.oneOrMore()._as("values"), s.IMPORTANT.optional()._as("important"), s.SEMICOLON.optional())
+	g.rule      ("MacroInvocation",  s.NAME._as("name"),   s.LP, s.Arguments.optional()._as("arguments"), s.RP)
 
 	# =========================================================================
 	# BLOCK STRUCTURE
@@ -178,14 +183,18 @@ def grammar(g=None):
 	g.rule("Statement",     s.CheckIndent._as("indent"), g.agroup(s.Assignment, s.MacroInvocation, s.COMMENT), s.EOL).disableFailMemoize()
 	g.rule("Block",         s.CheckIndent._as("indent"), g.agroup(s.PERCENTAGE, s.SelectionList)._as("selector"), s.COLON.optional(), s.EOL, s.Indent, s.Statement.zeroOrMore()._as("code"), s.Dedent).disableFailMemoize()
 
-	g.rule    ("SpecialDeclaration",   s.CheckIndent, s.SPECIAL_NAME, s.SPECIAL_FILTER.optional(), s.Parameters.optional(), s.COLON)
-	g.rule    ("SpecialBlock",         s.CheckIndent, s.SpecialDeclaration, s.EOL, s.Indent, s.Statement.zeroOrMore(), s.Dedent).disableFailMemoize()
+	g.rule    ("MacroDeclaration", s.MACRO, s.NAME._as("name"), s.Parameters.optional()._as("parameters"), s.COLON.optional())
+	g.rule    ("MacroBlock",       s.CheckIndent._as("indent"), s.MacroDeclaration._as("type"), s.EOL, s.Indent, s.Statement.zeroOrMore()._as("code"), s.Dedent).disableFailMemoize()
+
+	# FIXME: The special declaration is a bit broken... should work better
+	g.rule    ("SpecialDeclaration",   s.SPECIAL_NAME._as("type"), s.SPECIAL_FILTER.optional()._as("filter"),  s.NAME.optional()._as("name"), s.Parameters.optional()._as("parameters"), s.COLON.optional())
+	g.rule    ("SpecialBlock",         s.CheckIndent._as("indent"), s.SpecialDeclaration._as("type"), s.EOL, s.Indent, s.Statement.zeroOrMore()._as("code"), s.Dedent).disableFailMemoize()
 
 	# =========================================================================
 	# AXIOM
 	# =========================================================================
 
-	g.group     ("Source",  g.agroup(s.Comment, s.Block, s.SpecialBlock, s.Declaration, s.Include).zeroOrMore())
+	g.group     ("Source",  g.agroup(s.Comment, s.Block, s.MacroBlock, s.SpecialBlock, s.Declaration, s.Include).zeroOrMore())
 	g.skip(s.SPACE)
 	g.axiom(s.Source)
 	return g
@@ -319,11 +328,16 @@ class Processor(AbstractProcessor):
 		"""Resets the state of the processor. To be called inbetween parses."""
 		self.result     = []
 		self.indent     = 0
-		self.variables  = {}
-		self._evaluated = {}
+		self.variables  = [{}]
+		self.units      = {}
+		self._evaluated = [{}]
 		self.scopes     = []
 		self._header    = None
 		self._footer    = None
+		self._mode      = None
+		self._macro     = None
+		self._macros    = {}
+		self._property  = None
 
 	# ==========================================================================
 	# EVALUATION
@@ -335,12 +349,25 @@ class Processor(AbstractProcessor):
 
 		- values are encoded as `('V', (value, unit))`
 		- operations are encoded as `('O', operator, lvalue, rvalue)`
+		- literals are encoded as `('c', string)`
 		"""
-		if e[0] == "V":
+		if isinstance(e,list) and not isinstance(e[0], str):
+			return [[self.evaluate(_, unit, name, resolve, prefix) for _ in e], "l"]
+		if e[0] == "L":
+			return [[self.evaluate(_, unit, name, resolve, prefix) for _ in e[1]], "L"]
+		elif e[0] == "V":
 			v = e[1]
+			u = v[1]
+			if u in self.units:
+				# FIXME: It might be better to pre-calculate the unit values
+				# as this surely adds quite a lot of CPU time
+				u    = self.evaluate(self.units[u])
+				v    = (v[0] * u[0], u[1])
 			if resolve and v[1] == "R":
 				# We have a reference
-				return self.resolve(v[0], propertyName=name, prefix=prefix)
+				n = v[0]
+				v = self.resolve(v[0], propertyName=name, prefix=prefix)
+				return v
 			if self.IsColorProperty(name) and v[1] == "S":
 				# We have a color name as a string in a color property, we expand it
 				return (self.ColorFromName(v[0]) or v[0], "C")
@@ -378,16 +405,20 @@ class Processor(AbstractProcessor):
 		else:
 			raise ProcessingException("Evaluate not implemented for: {0} in {1}".format(e, name))
 
-	def resolve( self, name, propertyName=None, prefix=None ):
-		if name not in self.variables:
+	def resolve( self, name, propertyName=None, prefix=None, depth=1 ):
+		level     = len(self.variables) - depth
+		variables = self.variables[level] if len(self.variables) >= depth else None
+		if not variables:
 			raise ProcessingException("Variable not defined: {0}".format(name))
+		elif name not in variables:
+			return self.resolve(name, propertyName, prefix, depth + 1)
 		else:
 			cname = name + (":" + propertyName if propertyName else "") + (":" + prefix if prefix else "")
-			if cname in self._evaluated:
-				return self._evaluated[cname]
+			if cname in self._evaluated[level]:
+				return self._evaluated[level][cname]
 			else:
-				v = self.evaluate(self.variables[name], name=propertyName, prefix=prefix)
-				self._evaluated[cname] = v
+				v = self.evaluate(variables[name], name=propertyName, prefix=prefix)
+				self._evaluated[level][cname] = v
 				return v
 
 	# ==========================================================================
@@ -442,7 +473,9 @@ class Processor(AbstractProcessor):
 		return ["V", value]
 
 	def onParameters( self, match ):
-		return [self.defaultProcess(match[0])] + [self.process(_[1]) for _ in self.process(match[1])]
+		a = self.defaultProcess(match[0])
+		b = self.defaultProcess(match[1])
+		return [self.defaultProcess(match[0])] + [_[1] for _ in b or []]
 
 	def onArguments( self, match ):
 		m0 = self.process(match[0])
@@ -494,6 +527,12 @@ class Processor(AbstractProcessor):
 				res = ["O", rop, ["O", op, a, b], c]
 		return res
 
+	def onExpressionList( self, match, head, tail=None ):
+		if tail:
+			return [["L", [head] + [_[1] for _ in tail]]]
+		else:
+			return [head]
+
 	def onAttribute( self, match, name, value ):
 		return "[{0}{1}{2}]".format(name, value[0] if value else "", value[1] if value else "")
 
@@ -562,22 +601,60 @@ class Processor(AbstractProcessor):
 		self._header = ",\n".join((self._selectionAsString(_) for _ in self.scopes[-1])) + " {"
 		return self._header
 
+	def onSpecialDeclaration( self, match, type, filter, name, parameters ):
+		return (type, filter, name, parameters)
+
+	def onMacroDeclaration( self, match, name, parameters ):
+		return ("@macro", None, name, parameters)
+
+	def onMacroInvocation( self, match, name, arguments ):
+		if self._mode == "macro":
+			self._macro.append(lambda: self.onMacroInvocation(match, name, arguments))
+			return None
+		if name not in self._macros:
+			raise Exception("Macro not defined: {0}, got {1}".format(name, self._macros.keys()))
+		params    = self._macros[name][0] or []
+		scope     = {}
+		arguments = arguments or []
+		assert len(arguments) <= len(params), "Too many arguments given to macro: {0}, {1} given, expecting {2}".format(name, arguments, params)
+		for i,a in enumerate(arguments):
+			scope[params[i]] = ["V", a]
+		self.variables.append(scope)
+		self._evaluated.append({})
+		for line in self._macros[name][1]:
+			line()
+		self.variables.pop()
+		self._evaluated.pop()
+
 	def onNumber( self, match, value, unit ):
 		value = float(value) if "." in value else int(value)
 		unit  = unit if unit else None
+		if unit == "%": value = value / 100.0
 		return (value, unit)
 
-	def onDeclaration( self, match, name, value ):
-		name = name
-		self.variables[name] = value
+	def onDeclaration( self, match, decorator, name, value ):
+		assert len(value) == 1
+		value = value[0]
+		self._mode = None
+		if not decorator:
+			name = name
+			self.variables[-1][name] = value
+		elif decorator == "@unit":
+			self.units[name] = value
+		else:
+			raise NotImplementedError
 		return None
 
 	def onAssignment( self, match, name, values, important ):
+		if self._mode == "macro":
+			self._macro.append(lambda: self.onAssignment(match, name, values, important))
+			return None
 		if self._header:
 			self._write(self._header)
 			self._header = None
 			self._footer = "}\n"
 		suffix = "!important" if important else ""
+		self._property = name
 		try:
 			evalues = [self._valueAsString(self.evaluate(_, name=name)) for _ in values]
 		except ProcessingException as e:
@@ -598,7 +675,13 @@ class Processor(AbstractProcessor):
 			return self._write("{0}: {1}{2};".format(name, " ".join(evalues), suffix), indent=1)
 
 	def onBlock( self, match, indent ):
-		if self._footer:
+		if indent == 0:
+			self._mode   = None
+		elif self._mode == "macro":
+			self._macro.append(lambda: self.onBlock(match, indent + self.indent))
+			return None
+		self.indent = indent
+		if self._footer and self._mode == None:
 			self._write(self._footer)
 			self._footer = None
 		while len(self.scopes) > indent:
@@ -606,6 +689,31 @@ class Processor(AbstractProcessor):
 		self.process(match["selector"])
 		self.process(match["code"])
 
+	def onMacroBlock( self, match, type, indent=None):
+		if indent == 0:
+			self._mode  = None
+		assert self._mode != "macro"
+		type, filter, name, params = type
+		self._mode   = "macro"
+		self._macro  = []
+		self._macros[name] = [params, self._macro]
+		self.process(match["code"])
+
+	def onSpecialBlock( self, match, type, indent=None):
+		if indent == 0:
+			self._mode  = None
+		elif self._mode == "macro":
+			self._macro.append(lambda: self.onSpecialBlock(indent, match, type))
+			return None
+		if self._footer and self._mode == None:
+			self._write(self._footer)
+			self._footer = None
+		type, filter, name, params = type
+		self._write("{0}{1} {2} {3} {{".format(type, filter or "", name or "", ", ".join(params) if params else  ""))
+		self._mode = None
+		self.process(match["code"])
+		self._footer = "}"
+
 	def onSource( self, match ):
 		result = self.defaultProcess(match)
 		if self._footer:
@@ -668,6 +776,8 @@ class Processor(AbstractProcessor):
 
 	def _mergeScopes( self, a, b):
 		# Merges the contents of scope A and B
+		if not b: return a
+		if not a: return b
 		return [self._mergeScopeUnit(a[i], b[i]) or None for i in range(len(a))]
 
 	def _scopeAsString( self, scope ):
@@ -686,7 +796,18 @@ class Processor(AbstractProcessor):
 
 	def _valueAsString( self, value ):
 		v, u = value ; u = u or ""
-		if   u == "C":
+		if   u == "L":
+			return ", ".join([self._valueAsString(_) for _ in v])
+		elif u == "l":
+			return " ".join([self._valueAsString(_) for _ in v])
+		elif u == "%":
+			v = 100.0 * v
+			d = int(v)
+			if v == d:
+				return "{0:d}%".format(d)
+			else:
+				return "{0:f}%".format(v)
+		elif   u == "C":
 			if type(v) in (str, unicode):
 				# If we have a string instead of a tuple, we pass it as-is
 				return v
@@ -711,16 +832,16 @@ class Processor(AbstractProcessor):
 			return v + u
 		elif type(v) == str:
 			# FIXME: Proper escaping
-			if self.RE_UNQUOTED.match(v):
+			if self._property != "content" and self.RE_UNQUOTED.match(v):
 				return "{0:s}".format(v,u)
 			else:
-				return "{0:s}".format(json.dumps(v),u)
+				return "{0:s}".format(json.dumps(v).replace("\u", "\\"),u)
 		elif type(v) == unicode:
 			# FIXME: Proper escaping
-			if self.RE_UNQUOTED.match(v):
+			if self._property != "content" and self.RE_UNQUOTED.match(v):
 				return "{0:s}".format(v,u)
 			else:
-				return "{0:s}".format(json.dumps(v),u)
+				return "{0:s}".format(json.dumps(v).replace("\u", "\\"),u)
 		else:
 			raise ProcessingException("Value string conversion not implemented: {0}".format(value))
 
@@ -750,6 +871,7 @@ def run(args):
 	)
 	oparser.add_argument("files", metavar="FILE", type=str, nargs='+', help='The .pcss files to parse')
 	oparser.add_argument("--report",  dest="report",  action="store_true", default=False)
+	oparser.add_argument("--verbose",  dest="verbose",  action="store_true", default=False)
 	oparser.add_argument("--output",  type=str,  dest="output", default=None)
 	# We create the parse and register the options
 	args = oparser.parse_args(args=args)
@@ -758,6 +880,7 @@ def run(args):
 	if not args.files:
 		sys.stderr.write(USAGE + "\n")
 	output = sys.stdout
+	# if args.verbose: getGrammar().setVerbose()
 	if args.output: output = open(args.output, "w")
 	for a in args.files:
 		if reporter: reporter.info("Processing: {0}".format(a))
